import urllib.parse
import asyncio
import multiprocessing
import logging

from zmq_legos.mdp import Worker as MDPWorker

from .process import LammpsProcess


class LammpsWorker:
    def __init__(self, stop_event, scheduler, command=None, num_workers=None, loop=None):
        self.logger = logging.getLogger(f'{self.__module__}.{self.__class__.__name__}')
        self.command = command
        self.num_workers = num_workers or multiprocessing.cpu_count()
        if self.num_workers > multiprocessing.cpu_count():
            raise ValueError('cannot have more workers than cpus')

        parsed = urllib.parse.urlparse(scheduler)
        self.mdp_worker = MDPWorker(
            stop_event,
            max_messages=self.num_workers,
            protocol=parsed.scheme, port=parsed.port, hostname=parsed.hostname,
            loop=loop)

    async def create(self):
        self._processes = []
        self.logger.info(f'starting {self.num_workers} lammps processes')
        for _ in range(self.num_workers):
            process = LammpsProcess(command=self.command)
            await process.create(self.mdp_worker.queued_messages, self.mdp_worker.completed_messages)
            self._processes.append(process)

    async def shutdown(self):
        self.logger.info(f'shutting down {self.num_workers} lammps processes')
        for process in self._processes:
            process.shutdown()
        await self.mdp_worker.disconnect()

    async def run(self):
        await self.mdp_worker.run(b'lammps.job')



# Not related to lammps the least bit but needed for sorting structures
from functools import reduce
from math import gcd

import numpy as np


def plane_from_miller_index(lattice, miller_index):
    if len(miller_index) != 3:
        raise ValueError('miller indicies must have 3 integers')

    for index in miller_index:
        if not isinstance(index, int):
            raise ValueError('All miller indicies must be integers')

    zero_indicies = [i for i, index in enumerate(miller_index) if index == 0]
    if len(zero_indicies) == 3:
        raise ValueError('Cannot have all three indicies at zero')

    non_zero_indicies = [i for i, index in enumerate(miller_index) if index != 0]
    common_divisor = reduce(gcd, [miller_index[i] for i in non_zero_indicies[1:]], miller_index[non_zero_indicies[0]])

    point = miller_index[non_zero_indicies[0]] * lattice.matrix[0] / common_divisor
    vectors = [lattice.matrix[i] for i in zero_indicies]

    for i in range(len(non_zero_indicies) - 1):
        i1, i2 = non_zero_indicies[i], non_zero_indicies[i+1]
        p1 = lattice.matrix[i1] * miller_index[i1] / common_divisor
        p2 = lattice.matrix[i2] * miller_index[i2] / common_divisor
        vectors.append(p1 - p2)

    normal = np.cross(vectors[0], vectors[1])
    normal = normal / np.linalg.norm(normal)
    return point, normal


def structure_to_neb_input(structure):
    return '\n'.join([
        '{}'.format(len(structure)),
        '\n'.join(['{} {} {} {}'.format(i+1, *site.coords) for i, site in enumerate(structure)])
    ])



import logging
import multiprocessing

import asyncio
import zmq.asyncio

# from .client import LammpsDistributedClient
from .worker import LammpsWorker
from .scheduler import LammpsMaster


MASTER_URI = "tcp://127.0.0.1:8555"


def init_logging():
    logging.basicConfig(level=logging.INFO)


def init_event_loop():
    loop = zmq.asyncio.ZMQEventLoop()
    asyncio.set_event_loop(loop)
    return loop


def init_scheduler(stop_event):
    try:
        init_logging()

        scheduler = LammpsMaster(stop_event, MASTER_URI, loop=init_event_loop())
        scheduler.run()
    except KeyboardInterrupt:
        stop_event.set()


def init_worker(stop_event):
    async def run_worker(worker):
        await worker.create()
        await worker.run()

    try:
        init_logging()
        loop = init_event_loop()
        worker = LammpsWorker(stop_event, MASTER_URI, num_workers=4, loop=loop)
        loop.run_until_complete(run_worker(worker))
    except KeyboardInterrupt:
        stop_event.set()


def main():
    try:
        stop_event = multiprocessing.Event()
        worker_process = multiprocessing.Process(target=init_worker, args=(stop_event,))
        scheduler_process = multiprocessing.Process(target=init_scheduler, args=(stop_event,))
        worker_process.start()
        scheduler_process.start()
        scheduler_process.join()
    except KeyboardInterrupt:
        exit(0)
    finally:
        stop_event.set()


if __name__ == "__main__":
    main()



import os
from collections import OrderedDict

from pymatgen.core import Specie, Element, Structure
from pymatgen.core.periodic_table import _pt_data as periodic_table
import numpy as np

from .core import LammpsBox, LammpsPotentials

class LammpsInput:
    def __init__(self, lammps_script, lammps_data, additional_files=None):
        self.lammps_data = lammps_data
        self.lammps_script = lammps_script
        self.additional_files = additional_files or []

    def write_input(self, output_dir, input_filename="lammps.in", make_dir=True):
        if make_dir and not os.path.exists(output_dir):
            os.makedirs(output_dir)

        self.lammps_data.write_file(os.path.join(output_dir, self.lammps_script['read_data']))
        self.lammps_script.write_file(os.path.join(output_dir, input_filename))

        for file_buffer, filename in self.additional_files:
            with open(os.path.join(output_dir, filename)) as f:
                f.write(file_buffer)


class LammpsScript(OrderedDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self):
        lines = ""
        for k1, v1 in self.items():
            if isinstance(v1, dict):
                v1 = v1.values()
            if isinstance(v1, list):
                for x in v1:
                    lines = "".join([lines, "{} ".format(k1)])
                    lines = "".join([lines, str(x), os.linesep])
            else:
                lines = "".join([lines, "{} ".format(k1)])
                lines = "".join([lines, " {}{}".format(str(v1), os.linesep)])
        return lines

    @property
    def log_filename(self):
        log = self.get('log', 'log.lammps')
        if log is None:
            log = "log.lammps"
        elif isinstance(log, (list, tuple)):
            log = log[-1]
        return str(log).split(' ')[0]

    @property
    def data_filenames(self):
        data = self.get('read_data', None)
        if data is None:
            return []
        elif isinstance(data, (list, tuple)):
            return [str(d).split()[0] for d in data]
        else:
            return [str(data).split()[0]]

    @property
    def dump_filename(self):
        dump = self.get('dump', None)
        if dump is None:
            return None
        elif isinstance(dump, (list, tuple)):
            dump = log[-1]

        dump = str(dump).split()
        if len(dump) < 4:
            raise ValueError('Invalid dump command: %s' % dump)
        return dump[4]

    def write_file(self, filename):
        with open(filename, 'w') as f:
            f.write(str(self))


class LammpsData:
    HEADER_TAGS = {
        'atoms', 'bonds', 'angles', 'dihedrals', 'impropers',
        'atom types', 'bond types', 'angle types', 'dihedral types', 'improper types',
        'extra bond per atom', 'extra angle per atom', 'extra dihedral per atom', 'extra improper per atom', 'extra special per atom',
        'ellipsoids', 'lines', 'triangles', 'bodies', 'xlo xhi', 'ylo yhi', 'zlo zhi', 'xy xz yz'
    }
    # Otherwise implied integer
    FLOAT_HEADER_TAGS = {'xlo xhi', 'ylo yhi', 'zlo zhi', 'xy xz yz'}
    REQUIRED_HEADER_TAGS = {'xlo xhi', 'ylo yhi', 'zlo zhi', 'atoms', 'atom types'}

    REQUIRED_SECTION_TAGS = {'Atoms', 'Masses'}
    CHECKED_SECTION_TAGS = {'Atoms', 'Pair Coeffs', 'PairIJ Coeffs', 'Bond Coeffs', 'Angle Coeffs', 'Dihedral Coeffs', 'Improper Coeffs'}
    SECTIONS_TAGS = {
        'Atoms', 'Velocities', 'Masses', 'Ellipsoids', 'Lines', 'Triangles', 'Bodies',
        'Bonds', 'Angles', 'Dihedrals', 'Impropers',
        'Pair Coeffs', 'PairIJ Coeffs', 'Bond Coeffs', 'Angle Coeffs', 'Dihedral Coeffs', 'Improper Coeffs',
        'BondBond Coeffs', 'BondAngle Coeffs', 'MiddleBondTorsion Coeffs', 'EndBondTorsion Coeffs',
        'AngleTorsion Coeffs', 'AngleAngleTorsion Coeffs', 'BondBond13 Coeffs', 'AngleAngle Coeffs'
    }


    def __init__(self, name, symbol_indicies, masses, atoms, lammps_box, potentials=None, velocities=None):
        self.name = name
        self.symbol_indicies = symbol_indicies
        self.masses = masses
        self.atoms = atoms
        self.lammps_box = lammps_box
        self.potentials = potentials
        self.velocities = velocities
        if self.potentials:
            self.potentials.symbol_indicies = self.symbol_indicies

    @classmethod
    def from_structure(cls, structure, potentials=None, include_charge=False, include_velocities=False):
        lammps_box = LammpsBox.from_lattice(structure.lattice)
        name = 'pymatgen autogenerated data file'
        symbol_indicies = {}
        masses = {}
        atoms = []
        velocities = None
        if include_velocities:
            velocities = []
        counter = 1
        for atom in structure:
            if atom.specie not in symbol_indicies:
                symbol_indicies[atom.specie] = counter
                counter += 1

            if isinstance(atom.specie, Specie):
                element = atom.specie.element
                charge = atom.specie.oxi_state
            elif isinstance(atom.specie, Element):
                element = atom.specie
                charge = 0.0

            if atom.specie not in masses:
                masses[atom.specie] = element.atomic_mass

            if include_velocities:
                velocities.append(atom.properties.get('velocity', [0, 0, 0]))

            atoms.append([atom.specie, charge, atom.coords])
        return cls(name, symbol_indicies, masses, atoms, lammps_box,
                   potentials=potentials, velocities=velocities)

    @classmethod
    def _parse_data_file(cls, filename):
        def is_header(line):
            for header in cls.HEADER_TAGS:
                if header in line:
                    data = line.replace(header, '').strip().split()
                    if header in cls.FLOAT_HEADER_TAGS:
                        value = list(map(float, data))
                    else:
                        value = int(data[0])
                    return header, value
            return None

        def is_section(line):
            for section in cls.SECTIONS_TAGS:
                if section in line:
                    return section
            return None

        def parseline(line):
            line = line.strip()
            comment = ''
            if '#' in line:
                comment_index = line.find('#')
                comment = line[comment_index+1:].strip()
                line = line[:comment_index]
            return line, comment

        def data_to_array(data):
            if len(data) == 0:
                return np.array([])

            # either int or floats
            is_integer = np.vectorize(lambda f:f.is_integer())
            row_types = is_integer(data[0])
            for row in data:
                row_types  = row_types * is_integer(row)
            names = ['f%d' % i for i in range(len(row_types))]
            formats = [np.int64 if _ else np.float64 for _ in row_types]
            return np.rec.array(data, dtype={'names': names, 'formats': formats})

        headers = {}
        sections = {}

        with open(filename) as f:
            description = f.readline()
            f.readline() # second line blank

            in_header = True
            current_section = [None, None, []]

            for line in f:
                line, comment = parseline(line)
                if not line:
                    continue # Blank line or comment

                if in_header:
                    header_found = is_header(line)
                    if header_found:
                        header, value = header_found
                        headers[header] = value
                    elif is_section(line):
                        in_header = False
                        current_section[0] = is_section(line)
                        current_section[1] = comment or None
                        current_section[2] = []
                    else:
                        raise ValueError('line: %s not recognized' % line)
                elif is_section(line):
                    if current_section[0]:
                        section, check, data = current_section
                        sections[section] = {'data': data_to_array(data), 'check': check}
                    current_section[0] = is_section(line)
                    current_section[1] = comment or None
                    current_section[2] = []
                else: # Reading data from section
                    current_section[2].append(tuple(map(float, line.split())))
            if current_section[0]:
                section, check, data = current_section
                sections[section] = {'data': data_to_array(data), 'check': check}
        return description, headers, sections

    @classmethod
    def _validate_data_file(cls, headers, sections):
        for required_header_tag in cls.REQUIRED_HEADER_TAGS:
            if required_header_tag not in headers:
                return 'Required header tag %s not included' % required_header_tag

        for required_section_tag in cls.REQUIRED_SECTION_TAGS:
            if required_section_tag not in sections:
                return 'Required section tag %s not included' % required_section_tag
        # Obviously can do more checks

    @classmethod
    def from_file(cls, filename, atom_style=None):
        description, headers, sections = cls._parse_data_file(filename)
        errors = cls._validate_data_file(headers, sections)
        if errors:
            raise ValueError('data file is invalid: {}'.format(errors))

        xlo, xhi = headers['xlo xhi']
        ylo, yhi = headers['ylo yhi']
        zlo, zhi = headers['zlo zhi']
        xy, xz, yz = headers.get('xy xz yz', [0, 0, 0])
        lammps_box = LammpsBox(xhi, yhi, zhi, xlo, ylo, zlo, xy, xz, yz)

        # Guess symbol from closest atomic mass (yes not great for isotopes)
        elements = np.array(
            [tuple([Element(element).atomic_mass, Element(element)]) for element in periodic_table],
            dtype={'names': ['mass', 'symbol'], 'formats': [np.float64, np.chararray]}
        )
        symbol_indicies = {}
        index_symbols = {} # Makes element lookup quicker
        masses = {}
        for index, atomic_mass, in sections['Masses']['data']:
            symbol = elements['symbol'][np.abs(elements['mass'] - atomic_mass).argmin()]
            symbol_indicies[Element(symbol)] = index
            index_symbols[index] = Element(symbol)
            masses[Element(symbol)] = atomic_mass

        # Default full format or use check format
        atoms = []
        sections['Atoms']['data'].sort(order='f0') # f0 is default name of first field
        for atom in sections['Atoms']['data']:
            if sections['Atoms'].get('check') == 'full':
                atom_type, charge, *position = atom['f2'], atom['f3'], atom['f4'], atom['f5'], atom['f6']
            else:
                index, mol, atom_type, charge, *position = atom

            element = index_symbols[atom_type]
            atoms.append([element, charge, position])

        velocities = None
        if 'Velocities' in sections:
            sections['Velocities']['data'].sort(order='f0') # f0 is default name of first field
            velocities = [[vx, vy, vz] for vx, vy, vz in sections['Velocities']['data'][['f1', 'f2', 'f3']]]

        # Get Potentials
        # TODO only gets pair potentials for now and no reason to keep str
        pair_potentials = {}
        if 'PairIJ Coeffs' in sections:
            for s1, s2, *parameters in sections['PairIJ Coeffs']['data']:
                pair_potentials[(index_symbols[s1], index_symbols[s2])] = ' '.join(list(map(str, parameters)))

        potentials = LammpsPotentials(pair_potentials, symbol_indicies)

        return cls(description, symbol_indicies, masses, atoms, lammps_box,
                   potentials=potentials, velocities=velocities)

    @property
    def structure(self):
        lattice = self.lammps_box.lattice
        species = []
        positions = []
        for specie, charge, coords in self.atoms:
            positions.append(coords)
            if isinstance(specie, Specie):
                symbol = specie.element.symbol
            else:
                symbol = specie.symbol

            if charge:
                species.append(Specie(symbol, charge))
            else:
                species.append(Element(symbol))

        site_properties = {}
        if self.velocities:
            site_properties['velocities'] = self.velocities

        return Structure(lattice, species, positions, coords_are_cartesian=True, site_properties=site_properties)


    def __str__(self):
        lammps_data_str = [
            '{}\n'.format(self.name),
            '{} atoms\n'.format(len(self.atoms)),
            '{} atom types\n'.format(len(self.symbol_indicies)),
            '{}\n'.format(str(self.lammps_box)),
            'Masses\n',
            '\n'.join(['{} {}'.format(self.symbol_indicies[specie], float(mass)) for specie, mass in self.masses.items()]) + '\n',
            'Atoms\n',
            '\n'.join(['{} {} {} {} {} {} {}'.format(i+1, 1, self.symbol_indicies[specie], charge, *coords) for i, (specie, charge, coords) in enumerate(self.atoms)]) + '\n',
        ]

        if self.velocities:
            lammps_data_str.extend([
                'Velocities\n',
                '\n'.join(['{} {} {} {}'.format(i+1, *velocity) for i, velocity in enumerate(self.velocities)]) + '\n'
            ])

        if self.potentials:
            lammps_data_str.append('{}\n'.format(str(self.potentials)))

        return '\n'.join(lammps_data_str)

    def write_file(self, filename):
        with open(filename, 'w') as f:
            f.write(str(self))



import re

import numpy as np
from pymatgen.core import Structure

from .core import LammpsBox
from .inputs import LammpsData


def fields_view(array, fields):
    return array.getfield(np.dtype(
        {name: array.dtype.fields[name] for name in fields}
    ))


class LammpsRun(object):
    """ Parse Lammps Run

    """
    def __init__(self, lammps_data, lammps_log=None, lammps_dump=None):
        # self.lammps_script would be nice to have as well
        self.lammps_data = LammpsData.from_file(lammps_data)
        self.lammps_log = LammpsLog(lammps_log) if lammps_log else None
        self.lammps_dump = LammpsDump(lammps_dump) if lammps_dump else None
        self._generate_maps()

    def _generate_maps(self):
        self._atom_index = []
        for atom in self.initial_structure:
            self._atom_index.append(atom.specie)

    def get_structure(self, index):
        if self.lammps_dump is None:
            raise ValueError('Requires lammps dump to get structures in md simulation')

        timestep = self.lammps_dump.trajectories[index]
        if any(p not in timestep['atoms'].dtype.names for p in ['x', 'y', 'z']):
            raise ValueError('Atom dumps must include x y z positions to get structures')

        lammps_box = LammpsBox(**timestep['box'])
        species = self._atom_index
        positions = np.array(fields_view(timestep['atoms'], ['x', 'y', 'z']).tolist())

        site_properties = {}
        if all(p in timestep['atoms'].dtype.names for p in ['vx', 'vy', 'vz']):
            site_properties['velocities'] = fields_view(timestep['atoms'], ['vx', 'vy', 'vz']).tolist()

        return Structure(lammps_box.lattice, species, positions, coords_are_cartesian=True, site_properties=site_properties)

    def get_forces(self, index):
        if self.lammps_dump is None:
            raise ValueError('Requires lammps dump to get forces in md simulation')
        return self.lammps_dump.get_forces(index)

    def get_stress(self, index):
        if self.lammps_log is None:
            raise ValueError('Requires lammps log to get stress in md simulation')
        return self.lammps_log.get_stress(index)

    def get_energy(self, index):
        if self.lammps_log is None:
            raise ValueError('Requires lammps log to get stress in md simulation')
        return self.lammps_log.get_energy(index)

    @property
    def final_structure(self):
        return self.get_structure(-1)

    @property
    def final_forces(self):
        return self.get_forces(-1)

    @property
    def final_stress(self):
        return self.get_stress(-1)

    @property
    def initial_structure(self):
        return self.lammps_data.structure


class LammpsDump(object):
    """
    Parse the lammps dump file to extract useful info about the system.
    """

    def __init__(self, filename):
        self.filename = filename
        self._parse_dump()

    @property
    def timesteps(self):
        return np.array([t['timestep'] for t in self.trajectories])

    def get_forces(self, index):
        timestep = self.trajectories[index]
        if any(p not in timestep['atoms'].dtype.names for p in ['fx', 'fy', 'fz']):
            raise ValueError('Atom dumps must include fx fy fz to get forces')

        return np.array(fields_view(timestep['atoms'], ['fx', 'fy', 'fz']).tolist())

    def _parse_dump(self):
        """
        parse dump file
        """
        self.trajectories = []

        with open(self.filename) as f:
            trajectory = {}
            while True:
                line = f.readline()
                if "ITEM: TIMESTEP" in line:
                    line = f.readline()
                    trajectory['timestep'] = int(line)
                elif "ITEM: NUMBER OF ATOMS" in line:
                    line = f.readline()
                    trajectory['natoms'] = int(line)
                elif "ITEM: BOX BOUNDS" in line:
                    # determine format
                    if "xy xz yz" in line: # triclinic format
                        xlo, xhi, xy = list(map(float, f.readline().split()))
                        ylo, yhi, xz = list(map(float, f.readline().split()))
                        zlo, zhi, yz = list(map(float, f.readline().split()))
                    else:
                        xlo, xhi = list(map(float, f.readline().split()))
                        ylo, yhi = list(map(float, f.readline().split()))
                        zlo, zhi = list(map(float, f.readline().split()))
                        xy, xz, yz = 0, 0, 0
                    trajectory['box'] = {
                        'xlo': xlo, 'xhi': xhi,
                        'ylo': ylo, 'yhi': yhi,
                        'zlo': zlo, 'zhi': zhi,
                        'xy': xy, 'xz': xz, 'yz': yz
                    }
                elif "ITEM: ATOMS" in line:
                    labels = line.split()[2:]
                    formats = [np.int64] * 2 + [np.float64] * (len(labels) - 2)

                    atom_items = []
                    for i in range(trajectory['natoms']):
                        line_data = f.readline().split()
                        line_data = [int(_) for _ in line_data[:2]] + [float(_) for _ in line_data[2:]]
                        atom_items.append(tuple(line_data))
                    trajectory['atoms'] = np.array(atom_items, dtype={'names': labels, 'formats': formats})
                    trajectory['atoms'] = np.sort(trajectory['atoms'], order='id')
                    self.trajectories.append(trajectory)
                    trajectory = {}
                else:
                    break


class LammpsLog(object):
    """
    Parser for LAMMPS log file.
    """

    def __init__(self, log_file="lammps.log"):
        """
        Args:
            log_file (string): path to the loag file
        """
        self.log_file = log_file
        self._parse_log()

    @property
    def timesteps(self):
        return self.thermo_data['step'].view(np.float)

    def get_stress(self, index):
        timestep = self.thermo_data[index]
        if any(p not in timestep.dtype.names for p in ['Pxy', 'Pxz', 'Pyz', 'Pxx', 'Pyy', 'Pzz']):
            raise ValueError('Atom dumps must include Pxy, Pxz, Pyz, Pxx, Pyy, Pzz to get stress')

        pxx = timestep['Pxx']
        pyy = timestep['Pyy']
        pzz = timestep['Pzz']
        pxy = timestep['Pxy']
        pxz = timestep['Pxz']
        pyz = timestep['Pyz']

        return np.array([
            [pxx, pxy, pxz],
            [pxy, pyy, pyz],
            [pxz, pyz, pzz]
        ])

    def get_energy(self, index):
        timestep = self.thermo_data[index]
        if 'TotEng' not in timestep.dtype.names:
            raise ValueError('Atom dumps mult include TotEng to get total energy')
        return float(timestep['TotEng'])

    def _parse_log(self):
        """
        Parse the log file for the thermodynamic data.
        Sets the thermodynamic data as a structured numpy array with field names
        taken from the the thermo_style command.
        """
        thermo_int_styles = {
            'step', 'elapsed', 'elaplong', 'spcpu',
            'part', 'atoms', 'nbuild', 'ndanger'
        }

        thermo_header = []
        thermo_types = []
        thermo_data = []
        inside_thermo_block = False
        read_thermo_header = False
        with open(self.log_file, 'r') as logfile:
            for line in logfile:
                # timestep, the unit depedns on the 'units' command
                time = re.search('timestep\s+([0-9]+)', line)
                if time and not thermo_data:
                    self.timestep = float(time.group(1))

                # total number md steps
                steps = re.search('run\s+([0-9]+)', line)
                if steps and not thermo_data:
                    self.nmdsteps = int(steps.group(1))

                # logging interval
                thermo = re.search('thermo\s+([0-9]+)', line)
                if thermo and not thermo_data:
                    self.interval = float(thermo.group(1))

                # thermodynamic data, set by the thermo_style command
                if "Memory usage per processor = " in line or \
                   "Per MPI rank memory allocation" in line:
                    inside_thermo_block = True
                    read_thermo_header = False
                elif inside_thermo_block:
                    if not read_thermo_header:
                        if len(thermo_header) == 0:
                            thermo_header = line.split()
                            thermo_types = [np.int if h.lower() in thermo_int_styles else np.float for h in thermo_header]
                        else:
                            if thermo_header != line.split():
                                raise ValueError('Cannot parse log file where thermo_style changes from one run to next. We suggest doing multiple seperate calculations')
                        read_thermo_header = True
                    elif "Loop time of " in line:
                        inside_thermo_block = False
                        read_thermo_header = False
                    else:
                        thermo_data.append(tuple(t(v) for t, v in zip(thermo_types, line.split())))
        thermo_data_dtype = np.dtype([(header, nptype) for header, nptype in zip(thermo_header, thermo_types)])
        self.thermo_data = np.array(thermo_data, dtype=thermo_data_dtype)



""" Example Config

{
   'master': 'tcp://localhost:8555'
}

"""
import json
import asyncio
import socket
import urllib.parse

import click
import zmq.asyncio

from . import cli
from ..calculator import LammpsWorker, LammpsMaster


def init_event_loop():
    loop = zmq.asyncio.ZMQEventLoop()
    asyncio.set_event_loop(loop)
    return loop

def normalize_uri(uri):
    parsed_uri = urllib.parse.urlparse(uri)
    hostname = socket.gethostbyname(parsed_uri.hostname)
    if parsed_uri.port:
        parsed_uri = parsed_uri._replace(netloc=f'{hostname}:{parsed_uri.port}')
    else:
        parsed_uri = parsed_uri._replace(netloc=f'{hostname}')
    return urllib.parse.urlunparse(parsed_uri)

@cli.command()
@click.option('-m', '--master')
@click.option('-n', '--num-workers', type=int)
@click.option('--command')
@click.option('-c', '--config', type=click.Path(exists=True))
def worker(master, num_workers, command, config):
    if config:
        with open(config) as f:
            config = json.load(f)
    else:
        config = {'master': master}
    master_uri = config.get('master', master)
    if master_uri is None:
        raise ValueError('must specify master uri')

    async def run_worker(worker):
        await worker.create()
        await worker.run()

    try:
        stop_event = asyncio.Event()
        loop = init_event_loop()
        worker = LammpsWorker(stop_event, normalize_uri(master_uri), num_workers=num_workers, command=command, loop=loop)
        loop.run_until_complete(run_worker(worker))
    except KeyboardInterrupt:
        stop_event.set()
        loop.run_until_complete(worker.shutdown())



@cli.command()
@click.option('-m', '--master')
@click.option('-c', '--config', type=click.Path(exists=True))
def master(master, config):
    if config:
        with open(config) as f:
            config = json.load(f)
    else:
        config = {'master': master}
    master_uri = config.get('master')
    if master_uri is None:
        raise ValueError('must specify master uri')

    try:
        stop_event = asyncio.Event()
        scheduler = LammpsMaster(stop_event, normalize_uri(master_uri), loop=init_event_loop())
        scheduler.run()
    except KeyboardInterrupt:
        stop_event.set()



# lammps scheduler (to be run as executable should be simple wrapper) easiest
# simple run function

import asyncio
import urllib.parse
import logging

from zmq_legos.mdp import Scheduler as MDPScheduler


class LammpsMaster:
    def __init__(self, stop_event, scheduler, loop=None):
        self.logger = logging.getLogger(f'{self.__module__}.{self.__class__.__name__}')
        parsed = urllib.parse.urlparse(scheduler)
        self.mdp_scheduler = MDPScheduler(
            stop_event,
            protocol=parsed.scheme, port=parsed.port, hostname=parsed.hostname,
            loop=loop)

    def run(self):
        self.mdp_scheduler.run()

    def disconnect(self):
        self.mdp_scheduler.disconnect()



import multiprocessing
import asyncio
import urllib.parse
import pickle
import uuid
import logging

from zmq_legos.mdp import Client as MDPClient

from .process import LammpsProcess


class LammpsLocalClient:
    def __init__(self, command=None, num_workers=None):
        self.command = command
        self.logger = logging.getLogger(f'{self.__module__}.{self.__class__.__name__}')
        self.num_workers = num_workers or multiprocessing.cpu_count()
        if num_workers > multiprocessing.cpu_count():
            raise ValueError('cannot have more workers than cpus')
        self.lammps_jobs = {}

    async def create(self):
        self._pending_queue = asyncio.Queue()
        self._completed_queue = asyncio.Queue()
        self._processes = []
        self.logger.info(f'creating {self.num_workers} lammps processes')
        for _ in range(self.num_workers):
            process = LammpsProcess(command=self.command)
            await process.create(self._pending_queue, self._completed_queue)
            self._processes.append(process)
        self._completed_jobs_task = asyncio.ensure_future(self._handle_completed())

    async def _handle_completed(self):
        while True:
            client_id, message = await self._completed_queue.get()
            lammps_job_output = pickle.loads(message[0])
            (self.lammps_jobs.pop(lammps_job_output['id'])).set_result(lammps_job_output)

    def shutdown(self):
        for process in self._processes:
            process.shutdown()

    async def submit(self, stdin, files=None, properties=None):
        lammps_job_input = {
            'id': uuid.uuid4().hex,
            'stdin': stdin,
            'files': files or {},
            'properties': properties or set()
        }
        future = asyncio.Future()
        self.lammps_jobs[lammps_job_input['id']] = future
        await self._pending_queue.put((b'client_id', [pickle.dumps(lammps_job_input)]))
        return future


class LammpsDistributedClient:
    def __init__(self, scheduler, loop=None):
        parsed = urllib.parse.urlparse(scheduler)
        self.mdp_client = MDPClient(protocol=parsed.scheme, port=parsed.port, hostname=parsed.hostname, loop=loop)
        self.logger = logging.getLogger(f'{self.__module__}.{self.__class__.__name__}')
        self.lammps_jobs = {}

    async def create(self):
        self._completed_jobs_task = asyncio.ensure_future(self._handle_completed())

    async def submit(self, stdin, files=None, properties=None):
        lammps_job_input = {
            'id': uuid.uuid4().hex,
            'stdin': stdin,
            'files': files or {},
            'properties': properties or set()
        }
        future = asyncio.Future()
        self.lammps_jobs[lammps_job_input['id']] = future
        await self.mdp_client.submit(b'lammps.job', [pickle.dumps(lammps_job_input)])
        self.logger.debug(f'lammps job {lammps_job_input["id"]} submitted')
        return future

    def shutdown(self):
        pass

    async def _handle_completed(self):
        while True:
            service, message = await self.mdp_client.get()
            lammps_job_output = pickle.loads(message[0])
            self.logger.debug(f'lammps job {lammps_job_output["id"]} completed')
            (self.lammps_jobs.pop(lammps_job_output['id'])).set_result(lammps_job_output)



import math

from pymatgen import Lattice


class LammpsBox:
    def __init__(self, xhi, yhi, zhi, xlo=0, ylo=0, zlo=0, xy=0, xz=0, yz=0, angles=None):
        self.xlo = xlo
        self.xhi = xhi
        self.ylo = ylo
        self.yhi = yhi
        self.zlo = zlo
        self.zhi = zhi
        self.xy = xy
        self.xz = xz
        self.yz = yz

    @classmethod
    def from_lattice(cls, lattice):
        a, b, c = lattice.abc
        alpha_rad, beta_rad, gamma_rad = list(map(math.radians, lattice.angles))
        lx = a
        xy = b * math.cos(gamma_rad)
        xz = c * math.cos(beta_rad)
        ly = math.sqrt(b**2 - xy**2)
        yz = (b * c * math.cos(alpha_rad) - xy * xz) / ly
        lz = math.sqrt(c**2 - xz**2 - yz**2)
        return cls(lx, ly, lz, xy=xy, xz=xz, yz=yz)

    @property
    def lattice(self):
        lx = self.xhi - self.xlo
        ly = self.yhi - self.ylo
        lz = self.zhi - self.zlo
        xy, xz, yz = self.xy, self.xz, self.yz
        a = lx
        b = math.sqrt(ly**2 + xy**2)
        c = math.sqrt(lz**2 + xz**2 + yz**2)
        alpha = math.degrees(math.acos((xy * xz + ly * yz) / (b * c)))
        beta = math.degrees(math.acos(xz / c))
        gamma = math.degrees(math.acos(xy / b))
        return Lattice.from_parameters(a, b, c, alpha, beta, gamma)

    def as_dict(self):
        return {
            'xlo': self.xlo, 'xhi': self.xhi,
            'ylo': self.ylo, 'yhi': self.yhi,
            'zlo': self.zlo, 'zhi': self.zhi,
            'xy': self.xy, 'xz': self.xz, 'yz': self.yz
        }

    def __str__(self):
        return (
            '{} {} xlo xhi\n'
            '{} {} ylo yhi\n'
            '{} {} zlo zhi\n'
            '{} {} {} xy xz yz'
        ).format(self.xlo, self.xhi, self.ylo, self.yhi, self.zlo, self.zhi,
                 self.xy, self.xz, self.yz)


class LammpsPotentials:
    def __init__(self, pair, symbol_indicies=None):
        self.symbol_indicies = symbol_indicies
        self.pair_parameters = pair

    def __str__(self):
        symbol_indicies = self.symbol_indicies
        if symbol_indicies is None: # Generate temporary symbol indicies
            counter = 1
            for (s1, s2) in self.pair_parameters:
                if s1 not in symbol_indicies:
                    symbol_indicies[s1] = counter
                    counter += 1
                if s2 not in symbol_indicies:
                    symbol_indicies[s2] = counter
                    counter += 1

        # Dumb enforcement that i < j
        def ordered_atom_type(a1_type, a2_type):
            if a1_type < a2_type:
                return "{} {}".format(a1_type, a2_type)
            return "{} {}".format(a2_type, a1_type)

        return '\n'.join([
            'PairIJ Coeffs\n',
            '\n'.join(['{} {}'.format(ordered_atom_type(symbol_indicies[s1], symbol_indicies[s2]), parameters) for (s1, s2), parameters in self.pair_parameters.items()]),
        ])



import subprocess
from collections import deque
from concurrent.futures import Future
import uuid

from .inputs import LammpsScript


class LammpsExecutor:
    """ An implementation that tries to run lammps as a calculator

    This is a good idea for lots of calculations that run in a short
    period of time.
    """
    def __init__(self, command, num_processes=1, cwd=None):
        self._num_processes = num_processes
        self.command = command or ['lammps']
        self.count = 0

        self._process = subprocess.Popen(self.command, cwd=cwd,
                                         stdin=subprocess.PIPE,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)

    def _submit(self, script):
        """submit task to pool of processes"""
        pass

    def submit(self, script):
        if not isinstance(script, LammpsScript):
            script = LammpsScript(script)
        script['clear'] = ''
        script['print'] = str(uuid.uuid4())



from collections import Counter
from scipy.stats import norm
import matplotlib.pyplot as plt
from pymatgen.util.plotting_utils import get_publication_quality_plot
from scipy import stats
import numpy as np


class RadialDistributionFunction(object):
    """
    Calculate the average radial distribution function for a given set of structures.
    """

    def __init__(self, structures, ngrid=101, rmax=10.0, cellrange=1, sigma=0.1,
                 species=None, reference_species=None):
        """
        Args:
            structures (list of pmg_structure objects): List of structure
                objects with the same composition. Allow for ensemble averaging.
            ngrid (int): Number of radial grid points.
            rmax (float): Maximum of radial grid (the minimum is always set zero).
            cellrange (int): Range of translational vector elements associated
                with supercell. Default is 1, i.e. including the adjecent image
                cells along all three directions.
            sigma (float): Smearing of a Gaussian function.
            species ([string]): A list of specie symbols of interest.
            reference_species ([string]): set this option along with 'species'
                parameter to compute pair radial distribution function.
                eg: species=["H"], reference_species=["O"] to compute
                    O-H pair distribution in a water MD simulation.
        """

        if ngrid < 1:
            raise ValueError("ngrid should be greater than 1")

        if sigma <= 0.0:
            raise ValueError("sigma should be > 0")

        lattice = structures[0].lattice
        indices = [j for j, site in enumerate(structures[0]) if site.specie.symbol in species]

        if len(indices) == 0:
            raise ValueError("Given species are not in the structure")

        ref_indices = indices
        if reference_species:
            ref_indices = [j for j, site in enumerate(structures[0])
                           if site.specie.symbol in reference_species]

        self.rho = float(len(indices)) / lattice.volume
        fcoords_list = []
        ref_fcoords_list = []

        for s in structures:
            all_fcoords = np.array(s.frac_coords)
            fcoords_list.append(all_fcoords[indices, :])
            ref_fcoords_list.append(all_fcoords[ref_indices, :])

        dr = rmax / (ngrid - 1)
        interval = np.linspace(0.0, rmax, ngrid)
        rdf = np.zeros((ngrid), dtype=np.double)
        dns = Counter()

        # generate the translational vectors
        r = np.arange(-cellrange, cellrange + 1)
        arange = r[:, None] * np.array([1, 0, 0])[None, :]
        brange = r[:, None] * np.array([0, 1, 0])[None, :]
        crange = r[:, None] * np.array([0, 0, 1])[None, :]
        images = arange[:, None, None] + brange[None, :, None] + crange[None, None, :]
        images = images.reshape((len(r) ** 3, 3))

        # find the zero image vector
        zd = np.sum(images ** 2, axis=1)
        indx0 = np.argmin(zd)

        for fcoords, ref_fcoords in zip(fcoords_list, ref_fcoords_list):
            dcf = fcoords[:, None, None, :] + images[None, None, :,
                                              :] - ref_fcoords[None, :, None, :]
            dcc = lattice.get_cartesian_coords(dcf)
            d2 = np.sum(dcc ** 2, axis=3)
            dists = [d2[u, v, j] ** 0.5 for u in range(len(indices)) for v in
                     range(len(ref_indices))
                     for j in range(len(r) ** 3) if u != v or j != indx0]
            dists = filter(lambda e: e < rmax + 1e-8, dists)
            r_indices = [int(dist / dr) for dist in dists]
            dns.update(r_indices)

        for indx, dn in dns.most_common(ngrid):
            if indx > len(interval) - 1: continue

            if indx == 0:
                ff = np.pi * dr ** 2
            else:
                ff = 4.0 * np.pi * interval[indx] ** 2

            rdf[:] += stats.norm.pdf(interval, interval[indx], sigma) * dn \
                      / float(len(ref_indices)) / ff / self.rho / len(
                fcoords_list)

        self.structures = structures
        self.rdf = rdf
        self.interval = interval
        self.cellrange = cellrange
        self.rmax = rmax
        self.ngrid = ngrid
        self.species = species
        self.dr = dr

    @property
    def coordination_number(self):
        """
        returns running coordination number

        Returns:
            numpy array
        """
        return np.cumsum(self.rdf * self.rho * 4.0 * np.pi * self.interval ** 2)

    def get_rdf_plot(self, label=None, xlim=[0.0, 8.0], ylim=[-0.005, 3.0]):
        """
        Plot the average RDF function.
        """

        if label is None:
            symbol_list = [e.symbol for e in
                           self.structures[0].composition.keys()]
            symbol_list = [symbol for symbol in symbol_list if
                           symbol in self.species]

            if len(symbol_list) == 1:
                label = symbol_list[0]
            else:
                label = "-".join(symbol_list)

        plt = get_publication_quality_plot(12, 8)
        plt.plot(self.interval, self.rdf, color="r", label=label, linewidth=4.0)
        plt.xlabel("$r$ ($\AA$)")
        plt.ylabel("$g(r)$")
        plt.legend(loc='upper right', fontsize=36)
        plt.xlim(xlim[0], xlim[1])
        plt.ylim(ylim[0], ylim[1])
        plt.tight_layout()

        return plt

    def export_rdf(self, filename):
        """
        Output RDF data to a csv file.

        Args:
            filename (str): Filename. Supported formats are csv and dat. If
                the extension is csv, a csv file is written. Otherwise,
                a dat format is assumed.
        """
        fmt = "csv" if filename.lower().endswith(".csv") else "dat"
        delimiter = ", " if fmt == "csv" else " "
        with open(filename, "wt") as f:
            if fmt == "dat":
                f.write("# ")
            f.write(delimiter.join(["r", "g(r)"]))
            f.write("\n")

            for r, gr in zip(self.interval, self.rdf):
                f.write(delimiter.join(["%s" % v for v in [r, gr]]))
                f.write("\n")



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

# To use a consistent encoding
from codecs import open
from os import path
import sys


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        if isinstance(self.pytest_args, str):
            self.pytest_args = shlex.split(self.pytest_args)
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

version = '0.2.2'
setup(
    name='pymatgen-lammps',
    version=version,
    description='A LAMMPS wrapper using pymatgen',
    long_description=long_description,
    url=None,
    author='Chris Ostrouchov',
    author_email='chris.ostrouchov+pymatgen-lammps@gmail.com',
    classifiers=[
        'Programming Language :: Python :: 3.6',
    ],
    cmdclass = {'test': PyTest},
    keywords='lammps pymatgen wrapper',
    download_url = 'https://gitlab.aves.io/costrouc/pymatgen-lammps/repository/archive.zip?ref=v%s' % version,
    packages=find_packages(exclude=['docs', 'tests', 'notebooks']),
    install_requires=[
        'pymatgen==2017.7.4',
        'zmq_legos',
        'click'
    ],
    package_data={'lammps': ['sets/*.json']},
    tests_require=['pytest'],
    entry_points={
        'console_scripts': [
            'pylammps=lammps.__main__:main'
        ]
    }
)



import os
import re
import asyncio
import shutil
import tempfile
import pickle
import logging
import time
import shlex

from ..output import LammpsDump, LammpsLog


class LammpsProcess:
    def __init__(self, command=None):
        self.directory = tempfile.mkdtemp()
        self.command = shlex.split(command or 'lammps')
        self.logger = logging.getLogger(f'{self.__module__}.{self.__class__.__name__}')
        if not shutil.which(self.command[0]): # simple test
            raise ValueError(f'lammps executable {self.command[0]} does not exist')

    async def create(self, pending_queue, completed_queue):
        self.process = await self.create_lammps_process()
        self.pending_queue = pending_queue
        self.completed_queue = completed_queue
        self._job_task = asyncio.ensure_future(self._handle_jobs())

    def shutdown(self):
        self.process.kill() # TODO: not very nice
        shutil.rmtree(self.directory)

    async def create_lammps_process(self):
        return await asyncio.create_subprocess_exec(
            *self.command, cwd=self.directory,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT)

    def _write_inputs(self, lammps_job_input):
        self.logger.debug(f'lammps job {lammps_job_input["id"]} writing stdin and files {lammps_job_input["files"].keys()}')
        for filename, content in lammps_job_input['files'].items():
            with open(os.path.join(self.directory, filename), 'w') as f:
                f.write(content)
        self.process.stdin.write((
            f'{lammps_job_input["stdin"]}'
            f'\nprint "====={lammps_job_input["id"]}====="\nclear\n'
        ).encode('utf-8'))
        self.process.stdin.write(b'\nprint "' + b'hack to force flush' * 500 + b'"\n')

    async def _monitor_job(self, lammps_job_output):
        lammps_job_buffer = []
        lammps_job_regex = re.compile(b"^={5}(.{32})={5}\n$")
        self.logger.debug(f'monitoring running lammps job {lammps_job_output["id"]}')
        async for line in self.process.stdout:
            match = lammps_job_regex.match(line)
            if match:
                if lammps_job_output['id'] != match.group(1).decode():
                    raise ValueError('job id does not match currently running job (should not happen)')
                lammps_job_output['stdout'] = b''.join(lammps_job_buffer)
                self.logger.debug(f'lammps job {lammps_job_output["id"]} completed')
                return True
            elif b'ERROR' in line:
                lammps_job_buffer.append(line)
                lammps_job_output['stdout'] = b''.join(lammps_job_buffer)
                self.logger.debug(f'lammps job {lammps_job_output["id"]} encountered error')
                raise ValueError('error executing script')
            elif b'hack to force flush' not in line:
                lammps_job_buffer.append(line)

    def _process_results(self, lammps_job_input, lammps_job_output):
        log_filename = 'log.lammps'
        dump_filename = None
        for line in lammps_job_input['stdin'].split('\n'):
            tokens = line.split()
            if len(tokens) == 0:
                continue
            if tokens[0] == 'log':
                log_filename = tokens[1]
            elif tokens[0] == 'dump':
                dump_filename = tokens[5]

        lammps_log = LammpsLog(os.path.join(self.directory, log_filename))
        if dump_filename is None and ({'forces'} & lammps_job_input['properties'] != set()):
            raise ValueError('requested properties require dump file')
        elif dump_filename:
            lammps_dump = LammpsDump(os.path.join(self.directory, dump_filename))

        self.logger.debug(f'lammps job {lammps_job_input["id"]} properties {lammps_job_input["properties"]} being collected')
        if 'stress' in lammps_job_input['properties']:
            lammps_job_output['results']['stress'] = lammps_log.get_stress(-1).tolist()
        if 'energy' in lammps_job_input['properties']:
            lammps_job_output['results']['energy'] = lammps_log.get_energy(-1)
        if 'forces' in lammps_job_input['properties']:
            lammps_job_output['results']['forces'] = lammps_dump.get_forces(-1).tolist()

    async def _handle_jobs(self):
        while True:
            lammps_job_buffer = []
            client_id, message = await self.pending_queue.get()
            lammps_job_input = pickle.loads(message[0]) # lammps_job_input {id, stdin, files, properties}
            lammps_job_output = {'id': lammps_job_input['id'], 'stdout': None, 'results': {}, 'error': None}
            try:
                start_time = time.perf_counter()
                self._write_inputs(lammps_job_input)
                self.logger.debug(f'lammps job {lammps_job_output["id"]} writing inputs {time.perf_counter() - start_time} [sec]')
                start_time = time.perf_counter()
                await self._monitor_job(lammps_job_output) # job error restart lammps process
                self.logger.debug(f'lammps job {lammps_job_output["id"]} completed in {time.perf_counter() - start_time} [sec]')
                start_time = time.perf_counter()
                self._process_results(lammps_job_input, lammps_job_output)
                self.logger.debug(f'lammps job {lammps_job_output["id"]} processing results {time.perf_counter() - start_time} [sec]')
            except ValueError as error:
                if 'error executing script' in error.message:
                    self.logger.warning('restarting lammps process')
                    self.process.kill()
                    self.process = await self.create_lammps_process()
                lammps_job_output['error'] = error.message
            except Exception as e: # debugging (remove later)
                print(e)
            await self.completed_queue.put((client_id, [pickle.dumps(lammps_job_output)]))
            self.pending_queue.task_done()



import os
import json
import random

from .inputs import LammpsInput, LammpsScript
from .utils import structure_to_neb_input

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))


class LammpsSet(LammpsInput):
    def __init__(self, config_filename, lammps_data, data_filename='initial.data'):
        with open(os.path.join(MODULE_DIR, 'sets', config_filename + ".json")) as f:
            lammps_script = json.load(f, object_pairs_hook=LammpsScript)
        lammps_script['read_data'] = data_filename
        super().__init__(lammps_script, lammps_data)


class StaticSet(LammpsSet):
    def __init__(self, lammps_data, user_lammps_settings=None, **kwargs):
        super().__init__('static', lammps_data, **kwargs)
        if user_lammps_settings:
            self.lammps_script.update(user_lammps_settings)


class RelaxSet(LammpsSet):
    def __init__(self, lammps_data, relax_box=True, user_lammps_settings=None, **kwargs):
        super().__init__('relax', lammps_data, **kwargs)
        if not relax_box:
            self.lammps_script.update([('fix', [])])
        if user_lammps_settings:
            self.lammps_script.update(user_lammps_settings)


class NEBSet(LammpsSet):
    def __init__(self, lammps_data, final_structure, user_lammps_settings=None, **kwargs):
        super().__init__('neb', lammps_data, **kwargs)
        if user_lammps_settings:
            self.lammps_script.update(user_lammps_settings)
        self.final_structure = final_structure

    def write_input(self, output_dir, input_filename="lammps.in", make_dir=True):
        super().write_input(output_dir=output_dir, input_filename=input_filename,
                            make_dir=make_dir)

        # Write final structure for NEB calculation (uses linear interpolation)
        # TODO: Works for now
        with open(os.path.join(output_dir, 'last.coords'), 'w') as f:
            f.write(structure_to_neb_input(self.final_structure))


class NVESet(LammpsSet):
    def __init__(self, lammps_data, initial_temp=298.17, user_lammps_settings=None, **kwargs):
        """ Make sure to initialize velocity for correct initial energy

        Example:

          velocity: [
             "all create 298 313159265 units box",
             "all zero linear units box"
          ]

        Initializes atoms to initial temperature of 298 and ensures
        zero momentum in the box.
        """
        super().__init__('nve', lammps_data, **kwargs)
        if initial_temp:
            self.lammps_script['velocity'][0] = 'all create {:.3f} {} units box'.format(initial_temp, random.randint(0, 10000000))
        else:
            self.lammps_script['velocity'] = []
        if user_lammps_settings:
            self.lammps_script.update(user_lammps_settings)


class NVTSet(NVESet):
    def __init__(self, lammps_data,
                 temp_start=298.17, temp_end=None, temp_damp=100.0,
                 user_lammps_settings=None, **kwargs):
        super().__init__(lammps_data, **kwargs)
        temp_end = temp_end or temp_start
        self.lammps_script['velocity'][0] = 'all create {:.3f} {} units box'.format(temp_start, random.randint(0, 10000000))
        self.lammps_script['fix'] = '1 all nvt temp {:.3f} {:.3f} {:.3f}'.format(temp_start, temp_end, temp_damp)
        if user_lammps_settings:
            self.lammps_script.update(user_lammps_settings)


class NPTSet(NVESet):
    def __init__(self, lammps_data,
                 temp_start=298.17, temp_end=None, temp_damp=100.0,
                 press_start=0.0, press_end=None, press_damp=1000.0,
                 user_lammps_settings=None, **kwargs):
        super().__init__(lammps_data, **kwargs)
        temp_end = temp_end or temp_start
        press_end = press_end or press_start
        self.lammps_script['velocity'][0] = 'all create {:.3f} {} units box'.format(temp_start, random.randint(0, 10000000))
        self.lammps_script['fix'] = '1 all npt temp {:.3f} {:.3f} {:.3f} iso {:.3f} {:.3f} {:.3f}'.format(
            temp_start, temp_end, temp_damp, press_start, press_end, press_damp)
        if user_lammps_settings:
            self.lammps_script.update(user_lammps_settings)


class NPHSet(NVESet):
    def __init__(self, lammps_data,
                 press_start=0.0, press_end=None, press_damp=1000.0,
                 initial_temp=None,
                 user_lammps_settings=None, **kwargs):
        super().__init__(lammps_data, **kwargs)
        press_end = press_end or press_start
        if initial_temp:
            self.lammps_script['velocity'][0] = 'all create {:.3f} {} units box'.format(initial_temp, random.randint(0, 10000000))
        else:
            self.lammps_script['velocity'] = []
        self.lammps_script['fix'] = '1 all nph iso {:.3f} {:.3f} {:.3f}'.format(press_start, press_end, press_damp)
        if user_lammps_settings:
            self.lammps_script.update(user_lammps_settings)



from .core import LammpsBox, LammpsPotentials
from .inputs import LammpsData, LammpsScript, LammpsInput
from .output import LammpsLog, LammpsDump, LammpsRun
from .sets import (
    LammpsSet,
    StaticSet, RelaxSet, NEBSet,
    NVESet, NVTSet, NPTSet, NPHSet
)
